// University of Washington Bothell Graphics Library
// Authors: Kelvin Sung, Steve Baer
// The accompanying library supports CSS Graphics courses taught at UW-Bothell
// See: http://courses.washington.edu/css450/
//      http://courses.washington.edu/css451/
///////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "uwbgl_Common.h"
#include "uwbgl_DrawHelper4.h"

/// Implementation of a "camera" that defines a viewing state of a 3D scene
/// 
/// This class records and stores the parameters for:
///   - a camera: position, lookat (target), and upvector
///   - view frustum: fov, near and far planes
/// this class has no capability when it comes to computing the trasnformations required
/// for View and Projection operations. It is expected an API-dependent (e.g. WindowHandler)
/// will access the camera and view information and compute/load the matrix processors.
class UWB_Camera
{
public:
  UWB_Camera();

  /// Draw a representation of this camera in 3D world coordinates
  /// \param lod [in] Level of detail to draw the camera representation. Low and Medium LOD
  ///        just draws the line from the position to the target (look at). High LOD draws
  ///        the entire view frustum
  /// \param draw_helper [in] used to draw the camera with a specific graphics API
  /// \param aspect_ratio [in] aspect ratio of a parent viewport
  void Draw( eLevelOfDetail lod, UWB_DrawHelper& draw_helper, float aspect_ratio ) const;

  /// Set up the camera's view frustum
  /// \param fov [in] Field of View in degrees
  /// \param near_plane [in] Distance from the camera position to the frustum near plane
  /// \param far_plane [in] Distance from the camera position to the frustum far plane
  ///                  This should always be greater than the near_plane value
  void SetViewFrustumInDegree(float fov, float near_plane, float far_plane);
  /// Get the camera's view frustum
  /// \param fov [out] Field of View in degrees
  /// \param near_plane [out] Distance from the camera position to the frustum near plane
  /// \param far_plane [out] Distance from the camera position to the frustum far plane
  ///                  This should always be greater than the near_plane value
  void GetViewFrustumInDegree(float& fov, float& near_plane, float& far_plane) const;

  ///
  /// Set the camera position in 3d world coordinates
  void SetCameraPosition(float x, float y, float z);
  void SetCameraPosition(vec3 position);
  vec3 GetCameraPosition() const;

  /// Set the camera 3d world location that the camer is focused
  /// on. This is also commonly called the target position
  void SetCameraLookAt(float x, float y, float z);
  void SetCameraLookAt(vec3 at);
  vec3 GetCameraLookAt() const;

  ///
  /// Set the vector that defines "up" for the camera
  void SetCameraUpVector(float x, float y, float z);
  void SetCameraUpVector(vec3 up_vector);
  vec3 GetCameraUpVector() const;

  ///
  /// Set the camera's field of view angle in degrees
  void SetViewVolumeFOVInDegree(float v);
  float GetViewVolumeFOVInDegree() const;
  float GetViewVolumeFOVInRadian() const;

  /// Set the distance between the camera position and the frustum
  /// near plane
  void SetViewVolumeNearPlane(float n);
  float GetViewVolumeNearPlane() const;

  /// Set the distance between the camera position and the frustum
  /// far plane
  void SetViewVolumeFarPlane(float f);
  float GetViewVolumeFarPlane() const;

  /// Rotating the Camera With Mouse
  void ResetRotation();
  void BeginMouseRotate(int x, int y);
  void MouseRotate(int x, int y);
  void GetMouseRotationsInRadian(float &x, vec3 &xRef, float &y, vec3 &yRef) const;

  /// Tracking the Camera With Mouse
  void BeginMouseTrack(int x, int y);
  void MouseTrack(int x, int y);

  /// Zooming the Camera (by moving camera position towards the look-at)
  void BeginMouseZoom(int x);
  void MouseZoom(int x);

protected:  
  // Camera parameters
  vec3  m_CameraPos, m_LookAt, m_UpVector;
  
  // Visible Volume (Frustum) 
  float m_ViewFOV; // in degrees
  float m_ViewNearPlane, m_ViewFarPlane;

  /// To support interactive rotating the camera, here instead of actually computing the rotated
  /// camera and at position, we store the "requested rotation values" and pre-concatenate rotation
  /// matrix to ViewMatrix.
  float m_XRotStart, m_YRotStart;
  float m_XRot, m_YRot;           // in degree

  /// To support mouse-click-drag for computing the drag distance.
  /// For Tracking and for Zooming.
  int m_TmpInitXPos, m_TmpInitYPos;
};